"""Methods for the analysis of large-scale deep learning matrices via linear operators.

Implements spectral analysis methods and linear operators from the GGN's
hierarchical decomposition presented in

- Papyan, V. (2020). Traces of class/cross-class structure pervade deep learning
  spectra. Journal of Machine Learning Research (JMLR),
  https://jmlr.org/papers/v21/20-933.html
"""
