# convert pdf generated by LaTeX into svg
echo "Automatic conversion won't work (scipy/torch logos will have artifacts)."
echo "Use inkscape and follow the steps in https://graphicdesign.stackexchange.com/a/144670"
